@@ -1,9 +1,9 @@
 import re
 from typing import Any, Dict, List, Union
 from docassemble.base.functions import DANav
-from docassemble.base.util import log, word, DADict, DAList, DAObject, DAFile, DAFileCollection, DAFileList, defined, value, pdf_concatenate, DAOrderedDict, action_button_html, include_docx_template, user_logged_in, user_info, action_argument, send_email, docx_concatenate, reconsider, get_config, space_to_underscore, LatitudeLongitude
+from docassemble.base.util import log, word, DADict, DAList, DAObject, DAFile, DAFileCollection, DAFileList, defined, value, pdf_concatenate, zip_file, DAOrderedDict, action_button_html, include_docx_template, user_logged_in, user_info, action_argument, send_email, docx_concatenate, reconsider, get_config, space_to_underscore, LatitudeLongitude, DAStaticFile
 
-__all__ = ['ALAddendumField', 'ALAddendumFieldDict', 'ALDocumentBundle', 'ALDocument', 'ALDocumentBundleDict','safeattr','label','key']
+__all__ = ['ALAddendumField', 'ALAddendumFieldDict', 'ALDocumentBundle', 'ALDocument', 'ALStaticDocument', 'ALDocumentBundleDict','safeattr','label','key']
 
 DEBUG_MODE = get_config('debug')
 
@@ -43,25 +43,34 @@ def html_safe_str(the_string: str) -> str:
   """
   return re.sub( r'[^A-Za-z0-9]+', '_', the_string )
 
-def table_row( title, view_file:DAFile, download_file:DAFile=None, view_icon:str="eye", download_icon:str="download") -> str:
+#def table_row( title, view_file:DAFile, download_file:DAFile=None, view_icon:str="eye", download_icon:str="download") -> str:
+def table_row( title:str, button_htmls:List[str] = []) -> str:
   """
-  Uses the provided DAFile/DAFileCollection objects to build the row of a table in HTML format that allows
-  you to both view and download an ALDocument.
+  Uses the provided title and list of button html strings to
+  return the row of an AL document-styled table in HTML format.
   """
-  if not download_file:
-    download_file = view_file
   html = '\n\t<tr>'
   # html += '\n\t\t<td><i class="fas fa-file"></i>&nbsp;&nbsp;</td>'
   # TODO: Need to replace with proper CSS
-  html += '\n\t\t<td><div><strong>' + title + '</strong></div></td>'
-  html += '\n\t\t<td>'
-  html += action_button_html( view_file.url_for(), label=word("View"), size="md", icon=view_icon, color="secondary" )
-  html += action_button_html( download_file.url_for(attachment=True), size="md", label=word("Download"), icon=download_icon, color="primary" )
-
+  html += '\n\t\t<td class="al_doc_title"><strong>' + title + '</strong></td>'
+  html += '\n\t\t<td class="al_buttons">'
+  for button in button_htmls:
+    html += button
+  html += '</td>'
   html += '\n\t</tr>'
 
   return html
 
+def view_button( file:DAFile, label:str = "View", icon:str = "eye" ) -> str:
+  return action_button_html( file.url_for(attachment=False), label=word(label), icon=icon, color="secondary", size="md", classname='al_view' )
+
+def download_button( file:DAFile, label:str = "Download", icon:str = "download" ) -> str:
+  return action_button_html( file.url_for(attachment=True), label=word(label), icon=icon, color="primary", size="md", classname='al_download' )
+
+def zip_button( file:DAFile, label:str = "Download zip", icon:str = "file-archive" ) -> str:
+  return action_button_html( file.url_for(attachment=False), label=word(label), icon=icon, color="primary", size="md", classname='al_zip' )
+
+
 class ALAddendumField(DAObject):
   """
   Object representing a single field and its attributes as related to whether
@@ -94,8 +103,11 @@ class ALAddendumField(DAObject):
       return self.value()
 
     # If trigger is not a boolean value, overflow value is the value that starts at the end of the safe value.
-    safe_text = self.safe_value(overflow_message = overflow_message, input_width=input_width, preserve_newlines=preserve_newlines)
     original_value = self.value_if_defined()
+    safe_text = self.safe_value(overflow_message = overflow_message, 
+                                input_width=input_width, 
+                                preserve_newlines=preserve_newlines, 
+                                _original_value = original_value)
     if isinstance(safe_text,str):
       # Always get rid of double newlines, for consistency with safe_value.
       value_to_process = re.sub(r"[\r\n]+|\r+|\n+",r"\n",original_value).rstrip()
@@ -113,7 +125,7 @@ class ALAddendumField(DAObject):
       return value_to_process[overflow_start:]
 
     # Do not subtract length of overflow message if this is a list of objects instead of a string
-    return self.value_if_defined()[self.overflow_trigger:]
+    return original_value[self.overflow_trigger:]
 
   def max_lines(self, input_width:int=80, overflow_message_length=0) -> int:
     """
@@ -129,14 +141,26 @@ class ALAddendumField(DAObject):
     """
     return self.value_if_defined()
 
-  def safe_value(self, overflow_message:str="", input_width:int=80, preserve_newlines:bool=False):
+  def safe_value(self, overflow_message:str="", input_width:int=80, preserve_newlines:bool=False, _original_value=None):
     """
     Try to return just the portion of the variable
     that is _shorter than_ the overflow trigger. Otherwise, return empty string.
+    Args:
+        overflow_message (str): A short message to go on the page where text is cutoff.
+        input_width (int): The width, in characters, of the input box. Defaults to 80.
+        preserve_newlines (bool): Determines whether newlines are preserved in the "safe" text.
+            Defaults to False, which means all newlines are removed. This allows more text to appear
+            before being sent to the addendum.
+        _original_value (Any): for speed reasons, you can provide the full text and just use this
+            method to determine if the overflow trigger is exceeded. If no _original_value is
+            provided, this method will determine it using the value_if_defined() method.
     """
 
     # Handle simplest case first
-    value = self.value_if_defined()
+    if _original_value:
+      value = _original_value
+    else:
+      value = self.value_if_defined()
     if isinstance(value, str) and len(value) <= self.overflow_trigger and (value.count('\r') + value.count('\n')) == 0:
       return value
 
@@ -352,12 +376,22 @@ class ALAddendumFieldDict(DAOrderedDict):
     If the "style" is set to overflow_only, only return the overflow values.
     """
     if style == 'overflow_only':
-      return [field for field in self.values() if defined(field.field_name) and len(field.overflow_value())]
+      return [field for field in self.values() if len(field.overflow_value())]
     else:
       return [field for field in self.values() if defined(field.field_name)]
 
   def overflow(self):
     return self.defined_fields(style='overflow_only')
+  
+  def has_overflow(self)->bool:
+    """Returns True if any defined field's length exceeds the overflow trigger.
+    Returns:
+      bool: True if at least 1 field has "overflow" content, False otherwise.
+    """
+    for field in self.values():
+      if field.overflow_value():
+        return True
+    return False      
 
   #def defined_sections(self):
   #  if self.style == 'overflow_only':
@@ -379,10 +413,10 @@ class DALazyAttribute(DAObject):
 
 class ALDocument(DADict):
   """
-  An opinionated collection of typically three attachment blocks:
-  1. The final version of a document, at typical key "final"
-  2. The preview version of a document, at typical key "preview"
-  3. An addendum of a document, at the attribute `addendum`
+  An opinionated dictionary of attachment blocks. Typically there are three:
+  1. The final version of a document with a signature. Use key="final" as an argument in a method.
+  2. The preview version of a document with no signature. Use key="preview" as an argument in a method.
+  3. An addendum of a document contained in the attribute `addendum`. E.g. `my_doc.addendum`.
 
   Each form that an interview generates will get its own ALDocument object.
 
@@ -396,13 +430,76 @@ class ALDocument(DADict):
   Required attributes:
     - filename: name used for output PDF
     - title: display name for the output PDF
-    - enabled
-    - has_addendum: The default value is False. Set to True if the document has overflow, like for a PDF template.
+    - enabled: if this document should be created. See examples.
 
   Optional attribute:
     - addendum: an attachment block
-    - overflow_fields
-
+    - overflow_fields: ALAddendumFieldDict instance. These values will be used to detect and handle overflow.
+    - has_addendum: Defaults to False. Set to True if the document could have overflow, like for a PDF template.
+  
+  Examples:
+  
+  Simple use where the document is always enabled and will have no addendum
+  ```
+  ---
+  objects:
+    - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc", enabled=True)
+  --- 
+  attachment:
+      variable name: my_doc[i]  # This will usually be "final" or "preview"
+      ...
+  ```
+  
+  Enable a document conditionally
+  ```
+  ---
+  # See that `enabled` is not defined here
+  objects:
+    - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc")
+    - affidavit_of_indigency: ALDocument.using(filename="affidavit-of-indigency.pdf", title="Affidavit of Indigency")
+  ---
+  # an example of an arbitrary condition
+  code: |
+    my_doc.enabled = condition1 and condition2
+    enable_my_doc_conditionally = True
+  ---
+  # A common example many interviews would encounter for an affidavit of indigency
+  code: |  
+    affidavit_of_indigency.enabled = ask_indigency_questions and is_indigent
+  ---
+  attachment:
+      variable name: my_doc[i]
+      ...
+  ---
+  attachment:
+      variable name: affidavit_of_indigency[i]
+      ...
+  ---
+  ```
+  
+  For a document that may need an addendum, you must specify this when the object is created
+  or in a mandatory code block. The addendum will only be triggered if the document has "overflow"
+  in one of the fields that you specify.
+  ```
+  ---
+  objects:
+    - my_doc: ALDocument.using(filename="myDoc.pdf", title="myDoc", enabled=True, has_addendum=True)
+  --- 
+  attachment:
+      variable name: my_doc[i]
+      ...
+  ---
+  generic object: ALDocument
+  attachment:
+    variable name: x.addendum
+    docx template file: docx_addendum.docx
+  ---
+  code: |
+    my_doc.overflow_fields['big_text_variable'].overflow_trigger = 640 # Characters 
+    my_doc.overflow_fields['big_text_variable'].label = "Big text label" # Optional - you may use in your addendum
+    my_doc.overflow_fields['list_of_objects_variable'].overflow_trigger = 4 # Items in the list
+    my_doc.overflow_fields.gathered = True      
+  ```
   """
   filename: str
   title: str
@@ -492,12 +589,12 @@ class ALDocument(DADict):
         return [self[key], self.addendum]
       else:
         return [self[key]]
-
+      
   def need_addendum(self) -> bool:
     return hasattr(self, 'has_addendum') and self.has_addendum and self.has_overflow()
 
   def has_overflow(self) -> bool:
-    return len(self.overflow()) > 0
+    return self.overflow_fields.has_overflow()
 
   def overflow(self):
     return self.overflow_fields.overflow()
@@ -522,6 +619,57 @@ class ALDocument(DADict):
       overflow_message = self.default_overflow_message
     return self.overflow_fields[field_name].overflow_value(overflow_message=overflow_message, preserve_newlines=preserve_newlines, input_width=input_width)
 
+class ALStaticDocument(DAStaticFile):
+  """A class that allows one-line initialization of static documents to include in an ALDocumentBundle.
+  
+  Note:
+      You should always place the static file within the /data/static folder of a package.
+      ALDocumentBundle relies on a publically accessible file. The /data/templates folder is private.
+      
+  Attributes:
+      filename(str): the path to the file within /data/static/.
+      title(str): The title that will display as a row when invoked with `download_list_html()` method
+                  of an ALDocumentBundle.                  
+  Examples:
+      Add a static PDF file to a document bundle.
+      .. code-block:: yaml
+        ---
+        objects:
+          - static_test: ALStaticDocument.using(title="Static Test", filename="static.pdf", enabled=True)
+        ---
+        objects: 
+          - bundle: ALDocumentBundle.using(elements=[static_test], filename="bundle", title="Documents to download now")
+          
+  Todo:
+      Handle files placed in /data/templates if that turns out to be useful. Likely by copying into
+      a DAFile with pdf_concatenate().
+  """
+  def init(self, *pargs, **kwargs):
+    super().init(*pargs, **kwargs)
+    self.has_addendum = False
+    self.auto_gather = False
+    self.gathered = True
+  
+  def __getitem__(self, key):
+    # This overrides the .get() method so that the 'final' and 'private' key always exist and
+    # point to the same file.
+    return self
+  
+  def as_list(self, key:str='final', refresh:bool=True) -> List[DAFile]:
+    return [self]
+  
+  def as_pdf(self, key:str='final', refresh:bool=True) -> DAStaticFile:
+    return pdf_concatenate(self)
+    if self._is_pdf():
+      return self
+    else:
+      return pdf_concatenate(self)
+  
+  def show(self, **kwargs):
+    # TODO: this explicit conversion shouldn't be needed
+    # Workaround for problem generating thumbnails without it
+    return pdf_concatenate(self).show(**kwargs)
+  
 class ALDocumentBundle(DAList):
   """
   DAList of ALDocuments or nested ALDocumentBundles.
@@ -583,6 +731,33 @@ class ALDocumentBundle(DAList):
     setattr(self.cache, safe_key, pdf)
     return pdf
 
+  def as_zip(self, key:str = 'final', refresh:bool = True, title:str = '') -> DAFile:
+    '''Returns a zip file containing the whole bundle'''
+    log_if_debug(f'Calling as_zip() for { str( self.title )}')
+
+    zip_key = f'{ space_to_underscore( key )}_zip'
+
+    # Speed up performance if can (docs say `zip_file` works like `pdf_concatenate`)
+    if hasattr(self.cache, zip_key):
+      log_if_debug(f'Returning cached version of { str( self.title )} zip')
+      return getattr(self.cache,  zip_key)
+
+    # strip out a possible '.pdf' ending then add '.zip'
+    zipname = self.filename
+    if zipname.endswith( ".pdf" ):
+      zipname = zipname[:-len( ".pdf" )]
+    docs = [doc.as_pdf(key=key, refresh=refresh) for doc in self.enabled_documents()]
+    zip = zip_file( docs, filename=f'{ zipname }.zip' )
+    if title == '':
+      zip.title = self.title
+    else:
+      zip.title = title
+    
+    setattr(self.cache, zip_key, zip)
+    log_if_debug(f'Stored {self.title} zip at {self.instanceName}.cache.{zip_key}')
+    
+    return zip
+  
   def preview(self, refresh:bool=True) -> DAFile:
     return self.as_pdf(key='preview', refresh=refresh)
 
@@ -638,7 +813,7 @@ class ALDocumentBundle(DAList):
       editable.append(doc.docx if hasattr(doc, 'docx') else doc.pdf)
     return editable
 
-  def download_list_html(self, key:str='final', format:str='pdf', view:bool=True, refresh:bool=True) -> str:
+  def download_list_html(self, key:str='final', format:str='pdf', view:bool=True, refresh:bool=True, include_zip:bool = True) -> str:
     """
     Returns string of a table to display a list
     of pdfs with 'view' and 'download' buttons.
@@ -656,11 +831,18 @@ class ALDocumentBundle(DAList):
 
     for doc in self:
       if doc.enabled:
-        the_file = doc.as_pdf() # should trigger cache
         if format=='docx':
-          html += table_row(doc.title, the_file, download_file=doc.as_docx(key=key))
+          doc_dwnld_button = download_button( doc.as_docx(key=key) )
         else:
-          html += table_row(doc.title, the_file, download_file=the_file)
+          doc_dwnld_button = download_button( doc.as_pdf(key=key) )
+        buttons = [ view_button(doc.as_pdf()), doc_dwnld_button ]
+        html += table_row( doc.title, buttons )
+    
+    # Add a zip file row if there's more than one doc
+    if len(self.enabled_documents()) > 1 and include_zip:
+      zip = self.as_zip()
+      html += table_row( zip.title, zip_button( zip ))
+      
     html += '\n</table>'
 
     # Discuss: Do we want a table with the ability to have a merged pdf row?
@@ -688,9 +870,9 @@ class ALDocumentBundle(DAList):
     include an editable (Word) copy of the file, iff it is available.
     """
     name = re.sub(r'[^A-Za-z0-9]+','_', self.instanceName)  # safe name for classes and ids
-    al_wants_editable_input_id = 'al_wants_editable_' + name
-    al_email_input_id = 'al_doc_email_' + name
-    al_send_button_id = "al_send_email_button_"+name
+    al_wants_editable_input_id = '_ignore_al_wants_editable_' + name
+    al_email_input_id = '_ignore_al_doc_email_' + name
+    al_send_button_id = "al_send_email_button_" + name
 
     javascript_string = "javascript:aldocument_send_action('" + \
       self.attr_name('send_email_action_event') + \
@@ -699,24 +881,28 @@ class ALDocumentBundle(DAList):
 
     return_str = '''
   <div class="al_send_bundle '''+name+'''" id="al_send_bundle_'''+name+'''" name="al_send_bundle_'''+name+'''">
-  <h4 id="al_doc_email_header">Get a copy of the documents in email</h4>
-  <div class="al_email_container">
-  <span class="al_email_address '''+name+''' form-group row da-field-container da-field-container-datatype-email">
-    <label for="'''+al_email_input_id+'''" class="al_doc_email col-form-label da-form-label datext-right">Email</label>
-    <input value="''' + (user_info().email if user_logged_in() else '') + '''" alt="Input box" class="form-control" type="email" name="'''+al_email_input_id+'''" id="'''+al_email_input_id+'''">
-  </span>''' + action_button_html(javascript_string, label="Send", icon="envelope", color="primary", size="md", classname="al_send_email_button", id_tag=al_send_button_id) + "\n" + '''
-    </div>
+    <h5 id="al_doc_email_header">Get a copy of the documents in email</h5> 
     '''
     if show_editable_checkbox:
       return_str += '''
-    <div class="form-check-container"><div class="form-check">
-    <input class="form-check-input" type="checkbox" class="al_wants_editable" id="'''+al_wants_editable_input_id+'''">
-    <label class="al_wants_editable form-check-label" for="'''+al_wants_editable_input_id+'''">'''\
-      + word("Include an editable copy") + '''
-    </label>
+    <div class="form-check-container">
+      <div class="form-check">
+        <input class="form-check-input" type="checkbox" class="al_wants_editable" id="'''+al_wants_editable_input_id+'''">
+        <label class="al_wants_editable form-check-label" for="'''+al_wants_editable_input_id+'''">'''\
+          + word("Include an editable copy") + '''
+        </label>
+      </div>
+    </div>
+  '''
+    return_str += '''
+  <div class="al_email_container">
+    <span class="al_email_address '''+name+''' form-group row da-field-container da-field-container-datatype-email">
+      <label for="'''+al_email_input_id+'''" class="al_doc_email col-form-label da-form-label datext-right">Email</label>
+      <input value="''' + (user_info().email if user_logged_in() else '') + '''" alt="Input box" class="form-control" type="email" size="35" name="'''+al_email_input_id+'''" id="'''+al_email_input_id+'''">
+    </span>''' + action_button_html(javascript_string, label="Send", icon="envelope", color="primary", size="md", classname="al_send_email_button", id_tag=al_send_button_id) + "\n" + '''
   </div>
   '''
-    return_str += "</div>"
+    return_str += '</div>'  # al_send_bundle
     return return_str
 
   def send_email(self, to:any=None, key:str='final', editable:bool=False, template:any=None, **kwargs) -> bool:
